# type: ignore
from os import getenv
from typing import Union

from ._constants import (
    ACCOUNT_NAME_ENV_VAR,
    APP_KEY_ENV_VAR,
    APP_TOKEN_ENV_VAR,
    DEFAULT_RAISE_FOR_STATUS,
    DEFAULT_RETRY_ATTEMPTS,
    DEFAULT_RETRY_BACKOFF_EXPONENTIAL,
    DEFAULT_RETRY_BACKOFF_MAX,
    DEFAULT_RETRY_BACKOFF_MIN,
    DEFAULT_RETRY_LOGS,
    DEFAULT_RETRY_STATUSES,
    DEFAULT_TIMEOUT,
    RAISE_FOR_STATUS_ENV_VAR,
    RETRY_ATTEMPTS_ENV_VAR,
    RETRY_BACKOFF_EXPONENTIAL_ENV_VAR,
    RETRY_BACKOFF_MAX_ENV_VAR,
    RETRY_BACKOFF_MIN_ENV_VAR,
    RETRY_LOGS_ENV_VAR,
    RETRY_STATUSES_ENV_VAR,
    TIMEOUT_ENV_VAR,
)
from ._types import IterableType, UndefinedType
from ._utils import UNDEFINED, is_nullish_str, is_undefined, str_to_bool


class Config:
    def __init__(
        self,
        account_name: Union[str, UndefinedType] = UNDEFINED,
        app_key: Union[str, UndefinedType] = UNDEFINED,
        app_token: Union[str, UndefinedType] = UNDEFINED,
        timeout: Union[float, int, None, UndefinedType] = UNDEFINED,
        retry_attempts: Union[int, UndefinedType] = UNDEFINED,
        retry_backoff_min: Union[float, int, UndefinedType] = UNDEFINED,
        retry_backoff_max: Union[float, int, UndefinedType] = UNDEFINED,
        retry_backoff_exponential: Union[bool, float, int, UndefinedType] = UNDEFINED,
        retry_statuses: Union[IterableType[int], UndefinedType] = UNDEFINED,
        retry_logs: Union[bool, UndefinedType] = UNDEFINED,
        raise_for_status: Union[bool, UndefinedType] = UNDEFINED,
    ) -> None:
        self._account_name = self._parse_account_name(account_name)
        self._app_key = self._parse_app_key(app_key)
        self._app_token = self._parse_app_token(app_token)
        self._timeout = self._parse_timeout(timeout)
        self._retry_attempts = self._parse_retry_attempts(retry_attempts)
        self._retry_backoff_min = self._parse_retry_backoff_min(retry_backoff_min)
        self._retry_backoff_max = self._parse_retry_backoff_max(retry_backoff_max)
        self._retry_backoff_exponential = self._parse_retry_backoff_exponential(
            retry_backoff_exponential,
        )
        self._retry_statuses = self._parse_retry_statuses(retry_statuses)
        self._retry_logs = self._parse_retry_logs(retry_logs)
        self._raise_for_status = self._parse_raise_for_status(raise_for_status)

        if self.get_retry_backoff_min() > self.get_retry_backoff_max():
            raise ValueError("Minimum backoff has to be lower than maximum backoff")

    def with_overrides(
        self,
        account_name: Union[str, UndefinedType] = UNDEFINED,
        app_key: Union[str, UndefinedType] = UNDEFINED,
        app_token: Union[str, UndefinedType] = UNDEFINED,
        timeout: Union[float, int, None, UndefinedType] = UNDEFINED,
        retry_attempts: Union[int, UndefinedType] = UNDEFINED,
        retry_backoff_min: Union[float, int, UndefinedType] = UNDEFINED,
        retry_backoff_max: Union[float, int, UndefinedType] = UNDEFINED,
        retry_backoff_exponential: Union[bool, int, float, UndefinedType] = UNDEFINED,
        retry_statuses: Union[IterableType[int], UndefinedType] = UNDEFINED,
        retry_logs: Union[bool, UndefinedType] = UNDEFINED,
        raise_for_status: Union[bool, UndefinedType] = UNDEFINED,
    ) -> "Config":
        return Config(
            account_name=(
                self._account_name if is_undefined(account_name) else account_name
            ),
            app_key=self._app_key if is_undefined(app_key) else app_key,
            app_token=self._app_token if is_undefined(app_token) else app_token,
            timeout=self._timeout if is_undefined(timeout) else timeout,
            retry_attempts=(
                self._retry_attempts if is_undefined(retry_attempts) else retry_attempts
            ),
            retry_backoff_min=(
                self._retry_backoff_min
                if is_undefined(retry_backoff_min)
                else retry_backoff_min
            ),
            retry_backoff_max=(
                self._retry_backoff_max
                if is_undefined(retry_backoff_max)
                else retry_backoff_max
            ),
            retry_backoff_exponential=(
                self._retry_backoff_exponential
                if is_undefined(retry_backoff_exponential)
                else retry_backoff_exponential
            ),
            retry_statuses=(
                self._retry_statuses if is_undefined(retry_statuses) else retry_statuses
            ),
            retry_logs=(self._retry_logs if is_undefined(retry_logs) else retry_logs),
            raise_for_status=(
                self._raise_for_status
                if is_undefined(raise_for_status)
                else raise_for_status
            ),
        )

    def get_account_name(self) -> str:
        if is_undefined(self._account_name):
            raise ValueError("Missing VTEX Account Name")

        return self._account_name

    def get_app_key(self) -> str:
        if is_undefined(self._app_key):
            raise ValueError("Missing VTEX APP Key")

        return self._app_key

    def get_app_token(self) -> str:
        if is_undefined(self._app_token):
            raise ValueError("Missing VTEX APP Token")

        return self._app_token

    def get_timeout(self) -> Union[float, None]:
        if is_undefined(self._timeout):
            return DEFAULT_TIMEOUT

        return self._timeout

    def get_retry_attempts(self) -> int:
        if is_undefined(self._retry_attempts):
            return DEFAULT_RETRY_ATTEMPTS

        return self._retry_attempts

    def get_retry_backoff_min(self) -> float:
        if is_undefined(self._retry_backoff_min):
            return DEFAULT_RETRY_BACKOFF_MIN

        return self._retry_backoff_min

    def get_retry_backoff_max(self) -> float:
        if is_undefined(self._retry_backoff_max):
            return DEFAULT_RETRY_BACKOFF_MAX

        return self._retry_backoff_max

    def get_retry_backoff_exponential(self) -> float:
        if is_undefined(self._retry_backoff_exponential):
            return DEFAULT_RETRY_BACKOFF_EXPONENTIAL

        return self._retry_backoff_exponential

    def get_retry_statuses(self) -> IterableType[int]:
        if is_undefined(self._retry_statuses):
            return DEFAULT_RETRY_STATUSES

        return self._retry_statuses

    def get_retry_logs(self) -> bool:
        if is_undefined(self._retry_logs):
            return DEFAULT_RETRY_LOGS

        return self._retry_logs

    def get_raise_for_status(self) -> bool:
        if is_undefined(self._raise_for_status):
            return DEFAULT_RAISE_FOR_STATUS

        return self._raise_for_status

    def _parse_account_name(
        self,
        account_name: Union[str, UndefinedType] = UNDEFINED,
    ) -> Union[str, UndefinedType]:
        if isinstance(account_name, str) and account_name:
            return account_name

        if is_undefined(account_name):
            env_account_name = getenv(ACCOUNT_NAME_ENV_VAR, UNDEFINED)

            if is_undefined(env_account_name) or env_account_name:
                return env_account_name

            raise ValueError(
                f"Invalid value for {ACCOUNT_NAME_ENV_VAR}: {env_account_name}",
            )

        raise ValueError(f"Invalid value for account_name: {account_name}")

    def _parse_app_key(
        self,
        app_key: Union[str, UndefinedType] = UNDEFINED,
    ) -> Union[str, UndefinedType]:
        if isinstance(app_key, str) and app_key:
            return app_key

        if is_undefined(app_key):
            env_app_key = getenv(APP_KEY_ENV_VAR, UNDEFINED)

            if is_undefined(env_app_key) or env_app_key:
                return env_app_key

            raise ValueError(f"Invalid value for {APP_KEY_ENV_VAR}: {env_app_key}")

        raise ValueError(f"Invalid value for app_key: {app_key}")

    def _parse_app_token(
        self,
        app_token: Union[str, UndefinedType] = UNDEFINED,
    ) -> Union[str, UndefinedType]:
        if isinstance(app_token, str) and app_token:
            return app_token

        if is_undefined(app_token):
            env_app_token = getenv(APP_TOKEN_ENV_VAR, UNDEFINED)

            if is_undefined(env_app_token) or env_app_token:
                return env_app_token

            raise ValueError(f"Invalid value for {APP_TOKEN_ENV_VAR}: {env_app_token}")

        raise ValueError(f"Invalid value for app_token: {app_token}")

    def _parse_timeout(
        self,
        timeout: Union[float, int, None, UndefinedType] = UNDEFINED,
    ) -> Union[float, None, UndefinedType]:
        if isinstance(timeout, (float, int)) and timeout > 0:
            return float(timeout)

        if timeout is None:
            return timeout

        if is_undefined(timeout):
            env_timeout = getenv(TIMEOUT_ENV_VAR, UNDEFINED)

            if is_undefined(env_timeout):
                return env_timeout

            if is_nullish_str(env_timeout):
                return None

            try:
                converted_value = float(env_timeout)

                if converted_value > 0:
                    return converted_value
            except ValueError:
                pass

            raise ValueError(f"Invalid value for {TIMEOUT_ENV_VAR}: {env_timeout}")

        raise ValueError(f"Invalid value for timeout: {timeout}")

    def _parse_retry_attempts(
        self,
        retry_attempts: Union[int, UndefinedType] = UNDEFINED,
    ) -> Union[int, UndefinedType]:
        if isinstance(retry_attempts, int) and retry_attempts >= 0:
            return retry_attempts

        if is_undefined(retry_attempts):
            env_retry_attempts = getenv(RETRY_ATTEMPTS_ENV_VAR, UNDEFINED)

            if is_undefined(env_retry_attempts):
                return env_retry_attempts

            try:
                converted_value = int(env_retry_attempts)

                if converted_value >= 0:
                    return converted_value
            except ValueError:
                pass

            raise ValueError(
                f"Invalid value for {RETRY_ATTEMPTS_ENV_VAR}: {env_retry_attempts}",
            )

        raise ValueError(f"Invalid value for retry_attempts: {retry_attempts}")

    def _parse_retry_backoff_min(
        self,
        retry_backoff_min: Union[float, int, UndefinedType] = UNDEFINED,
    ) -> Union[float, UndefinedType]:
        if isinstance(retry_backoff_min, (float, int)) and retry_backoff_min > 0:
            return float(retry_backoff_min)

        if is_undefined(retry_backoff_min):
            env_retry_backoff_min = getenv(RETRY_BACKOFF_MIN_ENV_VAR, UNDEFINED)

            if is_undefined(env_retry_backoff_min):
                return env_retry_backoff_min

            try:
                converted_value = float(env_retry_backoff_min)

                if converted_value > 0:
                    return converted_value
            except ValueError:
                pass

            raise ValueError(
                f"Invalid value for {RETRY_BACKOFF_MIN_ENV_VAR}: "
                f"{env_retry_backoff_min}",
            )

        raise ValueError(f"Invalid value for retry_backoff_min: {retry_backoff_min}")

    def _parse_retry_backoff_max(
        self,
        retry_backoff_max: Union[float, UndefinedType] = UNDEFINED,
    ) -> Union[float, UndefinedType]:
        if isinstance(retry_backoff_max, (float, int)) and retry_backoff_max > 0:
            return float(retry_backoff_max)

        if is_undefined(retry_backoff_max):
            env_retry_backoff_max = getenv(RETRY_BACKOFF_MAX_ENV_VAR, UNDEFINED)

            if is_undefined(env_retry_backoff_max):
                return env_retry_backoff_max

            try:
                converted_value = float(env_retry_backoff_max)

                if converted_value > 0:
                    return converted_value
            except ValueError:
                pass

            raise ValueError(
                f"Invalid value for {RETRY_BACKOFF_MAX_ENV_VAR}: "
                f"{env_retry_backoff_max}",
            )

        raise ValueError(f"Invalid value for retry_backoff_max: {retry_backoff_max}")

    def _parse_retry_backoff_exponential(
        self,
        retry_backoff_exponential: Union[bool, float, int, UndefinedType] = UNDEFINED,
    ) -> Union[float, UndefinedType]:
        if (
            not isinstance(retry_backoff_exponential, bool)
            and isinstance(retry_backoff_exponential, (float, int))
            and retry_backoff_exponential >= 1
        ):
            return float(retry_backoff_exponential)
        elif isinstance(retry_backoff_exponential, bool):
            return (
                DEFAULT_RETRY_BACKOFF_EXPONENTIAL if retry_backoff_exponential else 1.0
            )

        if is_undefined(retry_backoff_exponential):
            env_retry_backoff_exponential = getenv(
                RETRY_BACKOFF_EXPONENTIAL_ENV_VAR,
                UNDEFINED,
            )

            if is_undefined(env_retry_backoff_exponential):
                return env_retry_backoff_exponential

            try:
                converted_value = float(env_retry_backoff_exponential)

                if converted_value >= 1:
                    return converted_value
            except ValueError:
                pass

            try:
                converted_value = str_to_bool(env_retry_backoff_exponential)
                return DEFAULT_RETRY_BACKOFF_EXPONENTIAL if converted_value else 1.0
            except ValueError:
                pass

            raise ValueError(
                f"Invalid value for {RETRY_BACKOFF_EXPONENTIAL_ENV_VAR}: "
                f"{env_retry_backoff_exponential}",
            ) from None

        raise ValueError(
            f"Invalid value for retry_backoff_exponential: {retry_backoff_exponential}",
        )

    def _parse_retry_statuses(
        self,
        retry_statuses: Union[IterableType[int], UndefinedType] = UNDEFINED,
    ) -> Union[IterableType[int], UndefinedType]:
        if isinstance(retry_statuses, (list, set, tuple)) and all(
            isinstance(status, int) and 100 <= status <= 599
            for status in retry_statuses
        ):
            return retry_statuses

        if is_undefined(retry_statuses):
            env_retry_statuses = getenv(RETRY_STATUSES_ENV_VAR, UNDEFINED)

            if is_undefined(env_retry_statuses):
                return env_retry_statuses

            try:
                converted_values = [
                    int(status.strip())
                    for status in env_retry_statuses.split(",")
                    if status.strip()
                ]

                if all(100 <= value <= 599 for value in converted_values):
                    return converted_values
            except ValueError:
                pass

            raise ValueError(
                f"Invalid value for {RETRY_STATUSES_ENV_VAR}: {env_retry_statuses}",
            ) from None

        raise ValueError(f"Invalid value for retry_statuses: {retry_statuses}")

    def _parse_retry_logs(
        self,
        retry_logs: Union[bool, UndefinedType] = UNDEFINED,
    ) -> Union[bool, UndefinedType]:
        if isinstance(retry_logs, bool):
            return retry_logs

        if is_undefined(retry_logs):
            env_retry_logs = getenv(RETRY_LOGS_ENV_VAR, UNDEFINED)

            if is_undefined(env_retry_logs):
                return env_retry_logs

            try:
                return str_to_bool(env_retry_logs)
            except ValueError:
                raise ValueError(
                    f"Invalid value for {RETRY_LOGS_ENV_VAR}: {env_retry_logs}"
                ) from None

        raise ValueError(f"Invalid value for retry_logs: {retry_logs}")

    def _parse_raise_for_status(
        self,
        raise_for_status: Union[bool, UndefinedType] = UNDEFINED,
    ) -> Union[bool, UndefinedType]:
        if isinstance(raise_for_status, bool):
            return raise_for_status

        if is_undefined(raise_for_status):
            env_raise_for_status = getenv(RAISE_FOR_STATUS_ENV_VAR, UNDEFINED)

            if is_undefined(env_raise_for_status):
                return env_raise_for_status

            try:
                return str_to_bool(env_raise_for_status)
            except ValueError:
                raise ValueError(
                    f"Invalid value for {RAISE_FOR_STATUS_ENV_VAR}: "
                    f"{env_raise_for_status}"
                ) from None

        raise ValueError(f"Invalid value for raise_for_status: {raise_for_status}")



from http import HTTPStatus
from json import JSONDecodeError
from logging import WARNING, Logger
from typing import Any, Type, Union, cast
from urllib.parse import urljoin

from httpx import (
    Client,
    Headers,
    HTTPError,
    HTTPStatusError,
    Response,
)
from httpx._types import (
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestFiles,
)
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from .._config import Config  # type: ignore[attr-defined]
from .._constants import APP_KEY_HEADER, APP_TOKEN_HEADER
from .._dto import VTEXResponse, VTEXResponseType
from .._exceptions import VTEXRequestError, VTEXResponseError
from .._logging import get_child_logger, get_logger
from .._types import HTTPMethodType
from .._utils import redact_headers


class BaseAPI:
    """
    Base client for VTEX API.
    """

    def __init__(
        self,
        config: Union[Config, None] = None,
        logger: Union[Logger, None] = None,
    ) -> None:
        self._config = config or Config()
        self._logger = (
            get_child_logger(logger, type(self).__name__)
            if logger
            else get_logger(type(self).__name__)
        )

    def _request(
        self,
        method: HTTPMethodType,
        environment: str,
        endpoint: str,
        headers: Union[HeaderTypes, None] = None,
        cookies: Union[CookieTypes, None] = None,
        params: Union[QueryParamTypes, None] = None,
        json: Union[Any, None] = None,
        data: Union[RequestData, None] = None,
        content: Union[RequestContent, None] = None,
        files: Union[RequestFiles, None] = None,
        config: Union[Config, None] = None,
        response_class: Union[Type[VTEXResponseType], None] = None,
    ) -> VTEXResponseType:
        request_config = self._get_config(config=config)
        url = self._get_url(
            config=request_config,
            environment=environment,
            endpoint=endpoint,
        )
        headers = self._get_headers(config=request_config, headers=headers)
        retry_statuses = set(request_config.get_retry_statuses())

        @retry(
            stop=stop_after_attempt(
                max_attempt_number=request_config.get_retry_attempts() + 1,
            ),
            wait=wait_exponential(
                min=request_config.get_retry_backoff_min(),
                max=request_config.get_retry_backoff_max(),
                exp_base=request_config.get_retry_backoff_exponential(),
            ),
            retry=retry_if_exception_type(exception_types=HTTPError),
            before_sleep=before_sleep_log(
                logger=self._logger,
                log_level=WARNING,
                exc_info=True,
            )
            if request_config.get_retry_logs()
            else None,
            reraise=True,
        )
        def send_vtex_request() -> Response:
            with Client(timeout=request_config.get_timeout()) as client:
                response = client.request(
                    method.upper(),
                    url,
                    headers=headers,
                    cookies=cookies,
                    params=params,
                    json=json,
                    data=data,
                    content=content,
                    files=files,
                )

                if response.status_code in retry_statuses:
                    response.raise_for_status()

                return response

        try:
            response = send_vtex_request()
        except HTTPStatusError as exception:
            response = exception.response
        except HTTPError as exception:
            headers = redact_headers(dict(headers))

            details = {
                "exception": exception,
                "method": str(method).upper(),
                "url": str(url),
                "headers": headers,
            }

            self._logger.error(str(exception), extra=details, exc_info=True)

            raise VTEXRequestError(**details) from None  # type: ignore[arg-type]

        response.request.headers = headers = Headers(redact_headers(dict(headers)))
        response.headers = Headers(redact_headers(dict(response.headers)))

        self._raise_from_response(response=response, config=request_config)

        return cast(
            VTEXResponseType,
            (response_class or VTEXResponse).factory(response),
        )

    def _get_config(self, config: Union[Config, None]) -> Config:
        return config or self._config

    def _get_url(self, config: Config, environment: str, endpoint: str) -> str:
        return urljoin(
            f"https://{config.get_account_name()}.{environment}.com.br",
            endpoint,
        )

    def _get_headers(
        self,
        config: Config,
        headers: Union[HeaderTypes, None] = None,
    ) -> Headers:
        request_headers = Headers(headers=headers)

        request_headers[APP_KEY_HEADER] = config.get_app_key()
        request_headers[APP_TOKEN_HEADER] = config.get_app_token()

        request_headers["Content-Type"] = "application/json; charset=utf-8"
        request_headers["Accept"] = "application/json"

        return request_headers

    def _raise_from_response(self, response: Response, config: Config) -> None:
        if response.is_error and config.get_raise_for_status():
            try:
                data = response.json(strict=False)
            except JSONDecodeError:
                data = response.text or HTTPStatus(response.status_code).phrase

            details = {
                "method": str(response.request.method).upper(),
                "url": str(response.request.url),
                "request_headers": response.request.headers,
                "status": response.status_code,
                "data": data,
                "response_headers": response.headers,
            }

            if response.is_server_error:
                self._logger.error(data, extra=details)
            else:
                self._logger.warning(data, extra=details)

            raise VTEXResponseError(data, **details) from None



from datetime import datetime, timedelta, timezone, tzinfo
from re import compile
from typing import Any, Dict, Mapping, Union

from distutils.util import strtobool

from ._constants import APP_KEY_HEADER, APP_TOKEN_HEADER
from ._types import JSONType, UndefinedType

TO_SNAKE_CASE_STEP_1_PATTERN = compile(r"(.)([A-Z][a-z]+)")
TO_SNAKE_CASE_STEP_2_PATTERN = compile(r"([a-z0-9])([A-Z])")

UNDEFINED = UndefinedType()


def is_nullish_str(value: str) -> bool:
    return value.lower() in {"", "null", "none", "nil"}


def is_undefined(value: Any) -> bool:
    return isinstance(value, UndefinedType)


def exclude_undefined_values(obj: Dict[Any, Any]) -> Dict[Any, Any]:
    return {key: value for key, value in obj.items() if not is_undefined(value)}


def str_to_bool(value: str) -> bool:
    return bool(strtobool(value))


def to_snake_case(string: str) -> str:
    return TO_SNAKE_CASE_STEP_2_PATTERN.sub(
        r"\1_\2",
        TO_SNAKE_CASE_STEP_1_PATTERN.sub(r"\1_\2", string),
    ).lower()


def to_snake_case_deep(obj: JSONType) -> JSONType:
    if isinstance(obj, dict):
        return {
            (to_snake_case(key) if isinstance(key, str) else key): (
                to_snake_case_deep(value)
            )
            for key, value in obj.items()
        }

    if isinstance(obj, (list, set, tuple)):
        return type(obj)([to_snake_case_deep(element) for element in obj])

    return obj


def redact_headers(headers: Mapping[str, str]) -> Dict[str, str]:
    redacted_headers = {}

    for key, value in list(headers.items()):
        if key.lower() in {APP_KEY_HEADER.lower(), APP_TOKEN_HEADER.lower()}:
            redacted_headers[key] = "*" * 32
        else:
            redacted_headers[key] = value

    return redacted_headers


def now(use_tz: bool = True, tz: Union[tzinfo, None] = None) -> datetime:
    return datetime.now((tz or timezone.utc) if use_tz else None)


def three_years_ago(use_tz: bool = True, tz: Union[tzinfo, None] = None) -> datetime:
    current_datetime = now(use_tz=use_tz, tz=tz)

    return current_datetime.replace(
        year=current_datetime.year - 3,
        hour=0,
        minute=0,
        second=0,
        microsecond=0,
    ) - timedelta(days=1)



from functools import cached_property
from typing import Union

from ._api import (
    CatalogAPI,
    CustomAPI,
    LicenseManagerAPI,
    LogisticsAPI,
    MasterDataAPI,
    OrdersAPI,
    TransactionsAPI,
)
from ._config import Config  # type: ignore[attr-defined]
from ._logging import get_logger
from ._types import IterableType, UndefinedType
from ._utils import UNDEFINED


class VTEX:
    """
    Entrypoint for the VTEX SDK.
    From this class you can access all the APIs on VTEX
    """

    def __init__(
        self,
        account_name: Union[str, UndefinedType] = UNDEFINED,
        app_key: Union[str, UndefinedType] = UNDEFINED,
        app_token: Union[str, UndefinedType] = UNDEFINED,
        timeout: Union[float, None, UndefinedType] = UNDEFINED,
        retry_attempts: Union[int, UndefinedType] = UNDEFINED,
        retry_backoff_min: Union[float, UndefinedType] = UNDEFINED,
        retry_backoff_max: Union[float, UndefinedType] = UNDEFINED,
        retry_backoff_exponential: Union[bool, float, UndefinedType] = UNDEFINED,
        retry_statuses: Union[IterableType[int], UndefinedType] = UNDEFINED,
        retry_logs: Union[bool, UndefinedType] = UNDEFINED,
        raise_for_status: Union[bool, UndefinedType] = UNDEFINED,
    ) -> None:
        self._logger = get_logger("client")
        self._config = Config(
            account_name=account_name,
            app_key=app_key,
            app_token=app_token,
            timeout=timeout,
            retry_attempts=retry_attempts,
            retry_backoff_min=retry_backoff_min,
            retry_backoff_max=retry_backoff_max,
            retry_backoff_exponential=retry_backoff_exponential,
            retry_statuses=retry_statuses,
            retry_logs=retry_logs,
            raise_for_status=raise_for_status,
        )

    @cached_property
    def custom(self) -> CustomAPI:
        return CustomAPI(config=self._config, logger=self._logger)

    @cached_property
    def catalog(self) -> CatalogAPI:
        return CatalogAPI(config=self._config, logger=self._logger)

    @cached_property
    def license_manager(self) -> LicenseManagerAPI:
        return LicenseManagerAPI(config=self._config, logger=self._logger)

    @cached_property
    def logistics(self) -> LogisticsAPI:
        return LogisticsAPI(config=self._config, logger=self._logger)

    @cached_property
    def master_data(self) -> MasterDataAPI:
        return MasterDataAPI(config=self._config, logger=self._logger)

    @cached_property
    def orders(self) -> OrdersAPI:
        return OrdersAPI(config=self._config, logger=self._logger)

    @cached_property
    def transactions(self) -> TransactionsAPI:
        return TransactionsAPI(config=self._config, logger=self._logger)



from typing import Any, Iterable, Union

from .._constants import (
    MIN_PAGE_SIZE,
    SEARCH_DOCUMENTS_MAX_PAGE_SIZE,
    SEARCH_DOCUMENTS_START_PAGE,
)
from .._dto import VTEXPaginatedListResponse
from .._types import IterableType, OrderingDirectionType, UndefinedType
from .._utils import UNDEFINED, exclude_undefined_values, is_undefined
from .base import BaseAPI


class MasterDataAPI(BaseAPI):
    """
    Client for the Master Data API.
    https://developers.vtex.com/docs/api-reference/master-data-api-v2
    """

    ENVIRONMENT = "vtexcommercestable"

    def search_documents(
        self,
        entity_name: str,
        where: Union[str, UndefinedType] = UNDEFINED,
        fields: Union[IterableType[str], str, UndefinedType] = UNDEFINED,
        order_by_field: str = "id",
        order_by_direction: OrderingDirectionType = "DESC",
        page: int = SEARCH_DOCUMENTS_START_PAGE,
        page_size: int = SEARCH_DOCUMENTS_MAX_PAGE_SIZE,
        **kwargs: Any,
    ) -> VTEXPaginatedListResponse:
        params = {
            "_where": where,
            "_sort": f"{order_by_field} {order_by_direction.upper()}",
        }

        if is_undefined(fields):
            params["_fields"] = "all"
        elif not isinstance(fields, str) and isinstance(fields, Iterable):
            params["_fields"] = ",".join(fields)
        else:
            params["_fields"] = fields

        page = max(page, SEARCH_DOCUMENTS_START_PAGE)
        page_size = max(
            min(page_size, SEARCH_DOCUMENTS_MAX_PAGE_SIZE),
            MIN_PAGE_SIZE,
        )

        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint=f"/api/dataentities/{entity_name}/search",
            params=exclude_undefined_values(params),
            headers={
                "REST-Range": f"resources={(page - 1) * page_size}-{page * page_size}",
            },
            config=self._config.with_overrides(**kwargs),
            response_class=VTEXPaginatedListResponse,
        )



from .catalog import CatalogAPI
from .custom import CustomAPI
from .license_manager import LicenseManagerAPI
from .logistics import LogisticsAPI
from .master_data import MasterDataAPI
from .orders import OrdersAPI
from .transactions import TransactionsAPI



from dataclasses import asdict, dataclass
from json import JSONDecodeError
from math import ceil
from typing import Dict, TypeVar, Union

from httpx import Request, Response

from ._types import IterableType, JSONType
from ._utils import to_snake_case_deep

VTEXResponseType = TypeVar("VTEXResponseType", bound="VTEXResponse", covariant=True)


@dataclass
class VTEXRequest:
    request: Request
    method: str
    url: str
    headers: Dict[str, str]

    @classmethod
    def factory(cls, request: Request) -> "VTEXRequest":
        return cls(
            request=request,
            method=str(request.method).upper(),
            url=str(request.url),
            headers=dict(request.headers),
        )


@dataclass
class VTEXResponse:
    request: VTEXRequest
    response: Response
    data: JSONType
    status: int
    headers: Dict[str, str]

    @classmethod
    def factory(cls, response: Response) -> "VTEXResponse":
        try:
            data = response.json(strict=False)
        except JSONDecodeError:
            data = response.text

        return cls(
            request=VTEXRequest.factory(response.request),
            response=response,
            data=to_snake_case_deep(data),
            status=int(response.status_code),
            headers=dict(response.headers),
        )


@dataclass
class VTEXListResponse(VTEXResponse):
    items: IterableType[JSONType]

    @classmethod
    def factory(cls, response: Response) -> "VTEXListResponse":
        vtex_response = VTEXResponse.factory(response)
        data = vtex_response.data

        if isinstance(data, list):
            items = data
        elif isinstance(data, dict) and isinstance(data.get("list"), list):
            items = data["list"]
        elif isinstance(data, dict) and isinstance(data.get("items"), list):
            items = data["items"]
        else:
            raise ValueError(f"Not a valid list response: {data}")

        return cls(**asdict(vtex_response), items=items)


@dataclass
class VTEXPagination:
    total: int
    pages: int
    page_size: int
    page: int
    previous_page: Union[int, None]
    next_page: Union[int, None]

    @classmethod
    def factory(cls, vtex_list_response: VTEXListResponse) -> "VTEXPagination":
        data, headers = vtex_list_response.data, vtex_list_response.headers

        total, pages, page_size, page = -1, -1, -1, -1
        if isinstance(data, dict) and data.get("paging"):
            pagination = data["paging"]
            total = pagination.get("total")
            pages = pagination.get("pages")
            page_size = pagination.get("per_page")
            page = pagination.get("page") or pagination.get("current_page")
        elif "rest-content-range" in headers:
            pagination = headers["rest-content-range"].split(" ")[-1]
            item_range, total = pagination.split("/")
            start, end = item_range.split("-")
            start, end, total = int(start), int(end), int(total)
            page_size = end - start
            pages = ceil(total / page_size)
            page = end // page_size

        if all(field != -1 for field in {total, pages, page_size, page}):
            return cls(
                total=total,
                pages=pages,
                page_size=page_size,
                page=page,
                previous_page=page - 1 if page > 1 else None,
                next_page=page + 1 if page < pages else None,
            )

        raise ValueError(f"Not a valid paginated list response: {vtex_list_response}")


@dataclass
class VTEXPaginatedListResponse(VTEXListResponse):
    pagination: VTEXPagination

    @classmethod
    def factory(cls, response: Response) -> "VTEXPaginatedListResponse":
        vtex_list_response = VTEXListResponse.factory(response)

        return cls(
            **asdict(vtex_list_response),
            pagination=VTEXPagination.factory(vtex_list_response),
        )


@dataclass
class VTEXScroll:
    token: Union[str, None]

    @classmethod
    def factory(cls, vtex_list_response: VTEXListResponse) -> "VTEXScroll":
        return cls(token=None)


@dataclass
class VTEXScrollListResponse(VTEXListResponse):
    scroll: VTEXScroll

    @classmethod
    def factory(cls, response: Response) -> "VTEXScrollListResponse":
        vtex_list_response = VTEXListResponse.factory(response)

        return cls(
            **asdict(vtex_list_response),
            scroll=VTEXScroll.factory(vtex_list_response),
        )



from typing import Any, Union

from .._constants import (
    LIST_SKU_IDS_MAX_PAGE_SIZE,
    LIST_SKU_IDS_START_PAGE,
    MIN_PAGE_SIZE,
)
from .._dto import VTEXListResponse, VTEXResponse
from .._types import UndefinedType
from .._utils import UNDEFINED, exclude_undefined_values
from .base import BaseAPI


class CatalogAPI(BaseAPI):
    """
    Client for the Catalog API.
    https://developers.vtex.com/docs/api-reference/catalog-api
    """

    ENVIRONMENT = "vtexcommercestable"

    def list_sku_ids(
        self,
        page: int = LIST_SKU_IDS_START_PAGE,
        page_size: int = LIST_SKU_IDS_MAX_PAGE_SIZE,
        **kwargs: Any,
    ) -> VTEXListResponse:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint="/api/catalog_system/pvt/sku/stockkeepingunitids",
            params={
                "page": max(page, LIST_SKU_IDS_START_PAGE),
                "pagesize": max(
                    min(page_size, LIST_SKU_IDS_MAX_PAGE_SIZE),
                    MIN_PAGE_SIZE,
                ),
            },
            config=self._config.with_overrides(**kwargs),
            response_class=VTEXListResponse,
        )

    def get_sku_with_context(self, sku_id: int, **kwargs: Any) -> VTEXResponse:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint=f"/api/catalog_system/pvt/sku/stockkeepingunitbyid/{sku_id}",
            config=self._config.with_overrides(**kwargs),
            response_class=VTEXResponse,
        )

    def list_sellers(
        self,
        sales_channel: Union[int, UndefinedType] = UNDEFINED,
        seller_type: Union[int, UndefinedType] = UNDEFINED,
        is_better_scope: Union[bool, UndefinedType] = UNDEFINED,
        **kwargs: Any,
    ) -> VTEXListResponse:
        return self._request(
            method="GET",
            environment=self.ENVIRONMENT,
            endpoint="/api/catalog_system/pvt/seller/list",
            params=exclude_undefined_values({
                "sc": sales_channel,
                "sellerType": seller_type,
                "isBetterScope": is_better_scope,
            }),
            config=self._config.with_overrides(**kwargs),
            response_class=VTEXListResponse,
        )



# Vtex Auth
APP_KEY_HEADER = "X-VTEX-API-AppKey"
APP_TOKEN_HEADER = "X-VTEX-API-AppToken"  # noqa: S105

# Configs
ACCOUNT_NAME_ENV_VAR = "VTEX_ACCOUNT_NAME"
APP_KEY_ENV_VAR = "VTEX_APP_KEY"
APP_TOKEN_ENV_VAR = "VTEX_APP_TOKEN"  # noqa: S105
TIMEOUT_ENV_VAR = "VTEX_TIMEOUT"
RETRY_ATTEMPTS_ENV_VAR = "VTEX_RETRY_ATTEMPTS"
RETRY_BACKOFF_MIN_ENV_VAR = "VTEX_RETRY_BACKOFF_MIN"
RETRY_BACKOFF_MAX_ENV_VAR = "VTEX_RETRY_BACKOFF_MAX"
RETRY_BACKOFF_EXPONENTIAL_ENV_VAR = "VTEX_RETRY_BACKOFF_EXPONENTIAL"
RETRY_STATUSES_ENV_VAR = "VTEX_RETRY_STATUSES"
RETRY_LOGS_ENV_VAR = "VTEX_RETRY_LOGS"
RAISE_FOR_STATUS_ENV_VAR = "VTEX_RAISE_FOR_STATUS"

DEFAULT_TIMEOUT = 10
DEFAULT_RETRY_ATTEMPTS = 5
DEFAULT_RETRY_BACKOFF_MIN = 1.0
DEFAULT_RETRY_BACKOFF_MAX = 64.0
DEFAULT_RETRY_BACKOFF_EXPONENTIAL = 2.0
DEFAULT_RETRY_STATUSES = {
    401,
    407,
    408,
    409,
    425,
    429,
    500,
    502,
    503,
    504,
    507,
    511,
    525,
}
DEFAULT_RETRY_LOGS = True
DEFAULT_RAISE_FOR_STATUS = True

# Generics
MIN_PAGE_SIZE = 1

# Catalog
LIST_SKU_IDS_START_PAGE = 1
LIST_SKU_IDS_MAX_PAGE_SIZE = 1000

# Logistics
LIST_CARRIERS_START_PAGE = 1
LIST_CARRIERS_MAX_PAGE_SIZE = 1000

LIST_DOCKS_START_PAGE = 1
LIST_DOCKS_MAX_PAGE_SIZE = 1000

LIST_SHIPPING_POLICIES_START_PAGE = 1
LIST_SHIPPING_POLICIES_MAX_PAGE_SIZE = 1000

# Master Data
SEARCH_DOCUMENTS_START_PAGE = 1
SEARCH_DOCUMENTS_MAX_PAGE_SIZE = 100

# Orders
LIST_FEED_ORDERS_MAX_PAGE_SIZE = 10
LIST_ORDERS_START_PAGE = 1
LIST_ORDERS_MAX_PAGE = 30
LIST_ORDERS_MAX_PAGE_SIZE = 100

# Transactions
LIST_TRANSACTION_INTERACTIONS_RETRIES = 10



from typing import (
    List,
    Literal,
    Mapping,
    Set,
    Tuple,
    TypeVar,
    Union,
)


class UndefinedType:
    def __repr__(self) -> str:
        return "Undefined"

    def __str__(self) -> str:
        return repr(self)


IterableValueType = TypeVar("IterableValueType")
IterableType = Union[
    Tuple[IterableValueType, ...],
    Set[IterableValueType],
    List[IterableValueType],
]

OrderingDirectionType = Literal["ASC", "DESC", "asc", "desc"]

HTTPMethodType = Literal[
    "DELETE",
    "GET",
    "HEAD",
    "OPTIONS",
    "PATCH",
    "POST",
    "PUT",
    "delete",
    "get",
    "head",
    "options",
    "patch",
    "post",
    "put",
]

PrimitiveTypes = Union[None, bool, int, float, str]
JSONType = Union[PrimitiveTypes, IterableType["JSONType"], Mapping[str, "JSONType"]]



from . import _constants as VTEXConstants  # noqa: N812
from ._dto import (
    VTEXListResponse,
    VTEXPaginatedListResponse,
    VTEXResponse,
    VTEXScrollListResponse,
)
from ._exceptions import VTEXError, VTEXRequestError, VTEXResponseError
from ._vtex import VTEX

__all__ = [
    "VTEX",
    "VTEXConstants",
    "VTEXError",
    "VTEXListResponse",
    "VTEXPaginatedListResponse",
    "VTEXRequestError",
    "VTEXResponse",
    "VTEXResponseError",
    "VTEXScrollListResponse",
]


for name in __all__:
    locals()[name].__module__ = "vtex"
